import json

from ridt.base import Error
from ridt.base import SettingErrorMessage

from ridt.config import RIDTConfig


class ConfigFileParser:
    """The class which handles the configuration file parsing.

    This class handles the parsing of configuration files from disk, as well as
    the instantiation of the :class:`~.RIDTConfig` instances using the parsed
    data.

    Attributes
    ----------
    path : :obj:`str`
        The path to the file containing the configuration JSON.
    
    data : :obj:`dict`
        The parsed JSON, stored as a :obj:`dict`.

    """
    def __new__(cls, *args, **kwargs):
        instance = super(ConfigFileParser, cls).__new__(cls)
        instance.__init__(*args, **kwargs)
        return instance.settings

    def __init__(self, path: str):
        self.path = path
        self.__parse_file()
        self.settings = self.__instantiate_settings()

    def __parse_file(self):
        """The method which parses the configuration file.

        Returns
        -------
        None

        Raises
        ------
        :class:`~.ConfigFileParserJSONError`
            If the structure of the parsed JSON is invalid.
        
        :class:`~.ConfigFileParserOSError`
            If an :obj:`OSError` is raised when parsing the file.

        """

        try:
            with open(self.path, 'r') as f:
                self.data = json.loads(f.read())
        except json.JSONDecodeError as e:
            raise ConfigFileParserJSONError(self.path, e)
        except OSError as e:
            raise ConfigFileParserOSError(self.path, e)

    def __instantiate_settings(self):
        """The method which instantiates the :class:`~.RIDTConfig` instances
        using the parsed data.

        Returns
        -------
        :class:`~.RIDTConfig`
            The :class:`~.RIDTConfig` instance created using the parsed data.
        
        Raises
        ------
        :class:`~.ConfigFileParserValidationError`
            If any :class:`~.SettingErrorMessage` errors are raised when
            instantiating the :class:`~.RIDTConfig` class.

        """
        try:
            return RIDTConfig(self.data)
        except SettingErrorMessage as e:
            raise ConfigFileParserValidationError(self.path, e)


class ConfigFileParserOSError(Error):
    """The exception raised when a :class:`~.ConfigFileParser` instance
    catches an error when parsing a config file path, or instantiating the
    :class:`~.Settings` derived class.

    """
    def __init__(self, path: str, error: OSError):
        """The constructor for the :class:`ConfigFileParserOSError` class.

        """
        msg = f"While loading {path}, I encountered the following error:"\
              f"\n {error}"
        super().__init__(msg)


class ConfigFileParserJSONError(Error):
    """The exception raised when a :class:`~.ConfigFileParser` instance
    catches an error when parsing a the JSON in a string parsed from a file.

    """
    def __init__(self, path: str, error: json.JSONDecodeError):
        """The constructor for the :class:`ConfigFileParserJSONError` class.

        """
        msg = f"While loading {path}, I encountered the following error:"\
              f"\n {error}. Please verify the integrity of the JSON file."
        super().__init__(msg)


class ConfigFileParserValidationError(Error):
    """The exception raised when a :class:`~.ConfigFileParser` instance
    catches an error when parsing a the JSON in a string parsed from a file.

    """
    def __init__(self, path: str, error: SettingErrorMessage):
        """The constructor for the :class:`ConfigFileParserValidationError`
        class.

        """
        msg = f"While loading {path}, I encountered the following error:"\
              f"\n {error}."
        super().__init__(msg)




import unittest
import os
import shutil

from ridt.config.csvtoconfigfile import CSVToConfigFile


class ST13(unittest.TestCase):

    """System Test 13. Test the system is able to take
       CSV files of data defining sources/monitor points
       and convert them into JSON configuration files."""

    def setUp(self) -> None:

        """setUp method that instantiates the
           :class:`~.RIDTConfig` class and the
           :class:`~.ComputationalSpace` class."""

        self.config_file = "tests/systemtests/st13/config.json"
        self.csv_file = "tests/systemtests/st13/info.csv"
        self.output = "tests/systemtests/st13"

    def test_modes(self):
        pass


if __name__ == "__main__":
    unittest.main()



from datetime import datetime


class Error(Exception):
    """The base class from which all other Error classes inherit.

    """
    pass


class RIDTOSError(Error):
    """The exception raised when RIDT attempts to modify a file but is unable.

    """
    def __init__(self, error):
        """The constructor for the :class:`RIDTOSError` class

        Parameters
        ----------
        obj
            The object which is not a :class:`~datetime.datetime` object.

        """
        msg = f"The following error was encountered: \n{error}"
        super().__init__(msg)


class NullError(Error):
    """The exception class raised when an operation is attempted on a null
    object.

    """

    def __init__(self, msg):

        super().__init__(msg)


class DateTimeTypeError(Error):
    """The exception raised if a value is not a :class:`~datetime.datetime`
    object, when it was expected to be.

    """

    def __init__(self, obj):
        """The constructor for the :class:`DateTimeTypeError` class

        Parameters
        ----------
        obj
            The object which is not a :class:`~datetime.datetime` object.

        """
        msg = f"The passed value {obj} was not {datetime}"
        super().__init__(msg)


class DictTypeError(Error):
    """The exception raised if a value is not a :obj:`dict`, when it was
    expected to be.

    """

    def __init__(self, obj):
        """The constructor for the :class:`DictTypeError` class

        Parameters
        ----------
        obj
            The object which is not a :obj:`dict`.

        """
        msg = f"The passed value {obj} was not {dict}"
        super().__init__(msg)


class FloatTypeError(Error):
    """The exception raised if a value is not a :obj:`float`, when it was
    expected to be.

    """

    def __init__(self, obj):
        """The constructor for the :class:`FloatTypeError` class

        Parameters
        ----------
        obj
            The object which is not a :obj:`float`.

        """
        msg = f"The passed value {obj} was not {float}"
        super().__init__(msg)


class ListTypeError(Error):
    """The exception raised if a value is not a :obj:`list`, when it was
    expected to be.

    """

    def __init__(self, obj):
        """The constructor for the :class:`ListTypeError` class

        Parameters
        ----------
        obj
            The object which is not a :obj:`list`.

        """
        msg = f"The passed value {obj} was not {list}"
        super().__init__(msg)


class StringTypeError(Error):
    """The exception raised if a value is not a :obj:`str`, when it was
    expected to be.

    """

    def __init__(self, obj):
        """The constructor for the :class:`StringTypeError` class

        Parameters
        ----------
        obj
            The object which is not a :obj:`str`.

        """
        msg = f"The passed value {obj} was not {str}"
        super().__init__(msg)


class IntTypeError(Error):
    """The exception raised if a value is not a :obj:`int`, when it was
    expected to be.

    """

    def __init__(self, obj):
        """The constructor for the :class:`IntTypeError` class

        Parameters
        ----------
        obj
            The object which is not a :obj:`int`.

        """
        msg = f"The passed value {obj} was not {int}"
        super().__init__(msg)


class UUIDNotFoundError(Error):
    """The exception raised if a uuid is not cannot be found in a particular
    register.

    """

    def __init__(self, uuid: str):
        """The constructor for the :class:`UUIDNotFoundError` class

        Parameters
        ----------
        obj
            The uuid which cannot be found.

        """
        msg = f"The uuid {uuid} could not be found."
        super().__init__(msg)


class DimensionError(Error):
    """The exception raised if a value is not of the desired dimension.
    """

    def __init__(self, obj, dim_required: int):
        """The constructor for the :class:`DimensionError` class

        Parameters
        ----------
        obj
            The object which is not a :obj:`int`.

        """
        msg = f"The dimension of {obj} is not {dim_required}"
        super().__init__(msg)



from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name='ridt',
    version='1.1.6',
    author="Riskaware Ltd",
    packages=find_packages(),
    description="Rapid Indoor Diffusion Tool",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/riskaware-ltd/ridt",
    include_package_data=True,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        'click',
        "numpy==1.16.5 ; python_version<'3.8'",
        "numpy==1.19.3 ; python_version>='3.8'",
        'scipy',
        'matplotlib',
        'tqdm'
    ],
    entry_points='''
        [console_scripts]
        ridt=ridt.cli.ridt:ridt
    ''',
)



import warnings

from os.path import join

from numpy import ndarray

from ridt.config import RIDTConfig
from ridt.config import Units

from ridt.container import Domain

import matplotlib.pyplot as plt
import matplotlib.dates as mdates 


def xfmt(x,pos=None):
    """custom date formatting"""
    x = mdates.num2date(x)
    label = x.strftime('%H:%M:%S')
    label = label.lstrip('0')
    return label


class PointPlot:
    """The class that generates line plots over time.

    Attributes
    ----------
    settings : :class:`~.RIDTConfig`
        The settings for the run in question.

    domain : :class:`~.Domain`
        The instance of :class:`~.Domain` corresponding to :attr:`setting`.

    units : :class:`~.Units`
        The instance of :class:`~.Units` corresponding to :attr:`setting`.
    
    quantity: :obj:`str`
        The string id for the quantity stored in the data  store.
    
    config : :class:`~.PointPlots`
        The point plots settings object.

    outdir: :obj:`str`
        The path to the output directory for the run.

    """
    def __init__(self, settings: RIDTConfig, output_dir: str, quantity: str):
        """The :class:`~.PointPlot` constructor.

        Parameters
        ----------
        settings : :class:`~.RIDTConfig`
            The settings for the run in question.

        outdir : :obj:`str`
            The path to the output directory for the run.

        quantity : :obj:`str`
            The string id for the quantity stored in the data  store.

        """
        self.settings = settings
        self.output_dir = output_dir
        self.units = Units(settings)
        self.domain = Domain(self.settings)
        self.quantity = quantity
        self.config = self.settings.models.eddy_diffusion.points_plots

    def __call__(self, id: str, data: ndarray):
        """The call method used to plot some data.

        Parameters
        ----------
        id : :obj:`str`
            The id string of the plane being plotted.

        data : :class:`~numpy.ndarray`
            The array of values to be plot.

        """
        self.id = id 
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            self.plot(data)
        plt.tight_layout()
        self.save_fig()
        plt.clf()

    def plot(self, data: ndarray):
        """Calls various plotting functions.

        Parameters
        ----------
        data : :class:`~numpy.ndarray`
            The array of values to be plot.

        """
        plt.title(self.title())
        plt.xlabel(self.xlabel())
        plt.ylabel(self.ylabel())
        if self.config.scale == "logarithmic":
            plt.yscale("log")
        plt.locator_params(nbins=10, axis='x')
        plot = plt.plot(self.convert_times(), data, marker='.')

        return plot

    def save_fig(self):
        """Creates the formatted string file name and saves plot to disk.

        """
        plt.savefig(join(self.output_dir, f"{self.id}.png"))

    def title(self):
        """Creates formatted title string for the current plot.

        Returns
        -------
        :obj:`str`
            The title string.

        """
        return f"{self.quantity} vs time - {self.id}"

    def ylabel(self):
        """Creates the formatted y-axis label string.

        Returns
        -------
        :obj:`str`
            The x-axis label string.

        """
        return f"{self.quantity} ({getattr(self.units, f'{self.quantity}')})"

    def xlabel(self):
        """Creates the formatted x-axis label string.

        Returns
        -------
        :obj:`str`
            The y-axis label string.

        """
        return f"Time ({self.config.time_axis_units}) [decimal fractions]"
    
    def convert_times(self):
        """Converts the time axis into date_time ticks

        Returns
        -------
        :obj:`list`[:obj:`str`]
            The times as datetime strings.

        """
        if self.config.time_axis_units == 's':
            return [t for t in self.domain.time]
        elif self.config.time_axis_units == 'm':
            return [t / 60 for t in self.domain.time]
        elif self.config.time_axis_units == 'h':
            return [t / 3600 for t in self.domain.time]
        


import unittest
import json
import os

import numpy as np

from ridt.data import DataStoreWriter
from ridt.data import DataStore

from ridt.config import RIDTConfig


class TestDataStoreWriter(unittest.TestCase):

    """Unit tests for the :class:`~.TestDataStoreWriter` class."""

    def setUp(self) -> None:

        """setUp method which creates a dummy data :obj:`dict`
        which is used within the tests. Additionally it
        instantiates the :class:`~.ConfigFileWriter` class."""

        this_dir = os.path.dirname(os.path.abspath(__file__))
        self.out_dir = os.path.join(this_dir, "test_resources/datastorewriter")
        self.point_name = "point"
        self.line_name = "line"
        self.plane_name = "plane"

        with open(os.path.join(this_dir, "../../default/config.json")) as f:
            loaded_json = json.load(f)

        self.config = RIDTConfig(loaded_json)

        self.ds = DataStore()
        self.space = np.linspace(0, 10, 10)

        self.point_data = np.array(self.space)

        self.line_data = np.array(
            [self.space, self.space])

        self.plane_data = np.array(
            [[self.space, self.space], [self.space, self.space]])

        self.domain_data = np.array(
            [[[self.space, self.space], [self.space, self.space]],
             [[self.space, self.space], [self.space, self.space]]]
        )

    def tearDown(self) -> None:
        for item in self.list_dir:
            path = os.path.join(self.out_dir, item)
            os.remove(path)



import unittest

import os

import numpy as np

from ridt.config import ConfigFileParser

from ridt.equation import WellMixed


class ST15(unittest.TestCase):

    """System Test 15. Test the system can use the
       information in the parsed JSON configuration
       file to evaluate the specified Well Mixed
       model for all times and points in parameter
       space (as defined by the ranges in the configuration file)."""

    def setUp(self) -> None:

        """setUp method that instantiates the
           :class:`~.RIDTConfig` class and the
           :class:`~.ComputationalSpace` class."""

        this_dir = os.path.dirname(os.path.abspath(__file__))
        with ConfigFileParser(os.path.join(this_dir, "st15/config.json")) as cfp:
            self.c = cfp

        self.time_array = np.linspace(
            0, self.c.total_time, self.c.time_samples
        )

        self.wm = WellMixed(self.c)

    def test_verify(self):
        conc = self.wm(self.time_array)
        self.assertEqual(
            len(conc), len(self.time_array)
        )


if __name__ == "__main__":
    unittest.main()



from typing import Union

from scipy.integrate import cumtrapz
from numpy import ndarray

from ridt.config import RIDTConfig

from ridt.data import DataStore
from ridt.data import BatchDataStore


class Exposure:
    """The Exposure class

    The class which takes a :class:`~.DataStore` or :class:`~.BatchDataStore`
    and returns another instance of the same containing the computed 
    exposure.

    Attributes
    ----------
    setting : :class:`~.RIDTConfig`
        The settings for the run in question.
    
    data_store : Union[:class:`~.DataStore`, :class:`~.BatchDataStore`]
        The data store or batch data store of computed exposures.
    
    delta_t : :obj:`float`
        The time step for the current data set.

    """

    def __new__(cls, *args, **kwargs):
        instance = super(Exposure, cls).__new__(cls)
        instance.__init__(*args, **kwargs)
        return instance.data_store
    
    def __init__(self, setting: RIDTConfig, data_store: Union[DataStore, BatchDataStore]):
        """The Exposure class initialiser


        Parameters
        ----------
        setting : :class:`~.RIDTConfig`
            The settings for the run in question.
        
        data_store : Union[:class:`~.DataStore`, :class:`~.BatchDataStore`]
            The data store or batch data store to be analysed.
        
        """
        self.setting = setting
        self.delta_t = self.setting.total_time / self.setting.time_samples
        self.data_store = self.evaluate(data_store)
    
    @property
    def geometries(self):
        """:obj:`list` [:obj:`str`] : the list of geometries selected for
        evaluation in :attr:`setting`.

        """
        locations = self.setting.models.eddy_diffusion.monitor_locations
        return [g for g, e in locations.evaluate.items() if e]

    def compute(self, data: ndarray) -> ndarray:
        """Cumulative integral over the time axis.

        Parameters
        ----------
        data : :class:`~numpy.ndarray`
            The data array to integrate over.

        Returns
        -------
        :class:`~numpy.ndarray`
            The integrated array containing exposures.

        """
        return cumtrapz(data, dx=self.delta_t, axis=0, initial=0)

    def evaluate(self, data_store: Union[DataStore, BatchDataStore])\
            -> Union[DataStore, BatchDataStore]:
        """Iterates over all geometries and computes exposures.

        Parameters
        ----------
        data_store : Union[:class:`~.DataStore`, :class:`~.BatchDataStore`]
            The data store or batch data store to be analysed.

        Returns
        -------
        Union[:class:`~.DataStore`, :class:`~.BatchDataStore`]
            The computed exposures.

        Raises
        ------
        :obj:`TypeError`
            If the `data_store` is not a :class:`~.DataStore` or
            :class:`~.BatchDataStore`.

        """
        if isinstance(data_store, DataStore):
            rv = DataStore()
            for geometry in self.geometries:
                for name, data in getattr(data_store, geometry).items():
                    rv.add(geometry, name, self.compute(data))
            return rv
        elif isinstance(data_store, BatchDataStore):
            rv = BatchDataStore()
            for setting, store in data_store.items():
                rv.add_run(setting)
                rv[setting] = self.evaluate(data_store[setting])
            return rv
        else:
            raise TypeError(f"Expecting {DataStore} or {BatchDataStore}.")
    


import warnings

import numpy

from typing import List
from typing import Tuple 
from typing import Union

from itertools import product

from tqdm import tqdm

from numpy import ndarray
from numpy import array
from numpy import zeros
from numpy import exp
from numpy import log
from numpy import power
from numpy import pi
from numpy import square
from numpy import nanmean

from scipy.integrate import cumtrapz
from scipy.integrate import romberg

from ridt.config import RIDTConfig
from ridt.config import InstantaneousSource
from ridt.config import InfiniteDurationSource
from ridt.config import FixedDurationSource 

from ridt import bar_args

numpy.seterr(divide='ignore')
numpy.seterr(invalid='ignore')


Source = Union[InstantaneousSource, InfiniteDurationSource, FixedDurationSource]
Value = Union[ndarray, float]
FloatList = List[float]

MAX_IMAGE = 20


class EddyDiffusion:
    """The core Eddy Diffusion model class.

    This class provides all equation implementation of the Eddy Diffusion Model.

    For details of the mathematics implemented here, please see the user guide.

    Attributes
    ----------
    settings : :class:`~.RIDTConfig`
        The settings for the run in question.
    
    dim : :class:`~.ridt.config.ridtconfig.Dimensions`
        The dimensions settings object that define the bounds of the system.
    
    disc : :class:`~.SpatialSamples`
        The spatial samples settings object that define how each axis is
        discretised.

    volume : :obj:`float`
        The total volume of the system.
    
    shape : :obj:`Tuple`[:obj:`int`]
        The shape of the current grid.
    
    cumtrapz_kwargs : :obj:`dict`
        A dictionary of keyword args for the cumulative trapezoidal integration
        function. 
    
    diff_coeff : :obj:`float`
        The diffusion coefficient.
    
    modes : :obj:`List`[:obj:`str`]
        The different string ids for the source modes.
    
    x : :class:`~numpy.ndarray`
        The current x-axis meshgrid.

    y : :class:`~numpy.ndarray`
        The current y-axis meshgrid.

    
    z : :class:`~numpy.ndarray`
        The current z-axis meshgrid.
    
    t : :obj:`List`[:obj:`float`]
        The current time domain array.

    rv : :class:`~numpy.ndarray`
        The calculated concentration values.

    """

    def __init__(self, settings: RIDTConfig):
        """The :class:`EddyDiffusion` constructor.

        Parameters
        ----------
        settings : :class:`~.RIDTConfig`
            The settings for the run in question.
            
        """
        self.settings = settings
        self.dim = self.settings.dimensions
        self.disc = self.settings.spatial_samples
        self.volume = self.dim.x * self.dim.y * self.dim.z
        self.cumtrapz_kwargs = {
            "axis": 0,
            "initial": 0,
            "dx": self.settings.total_time / self.settings.time_samples
        }
        self.diff_coeff = self.diffusion_coefficient()
        self.modes = ["instantaneous", "infinite_duration", "fixed_duration"]

    def __call__(self, x: ndarray, y: ndarray,  z: ndarray, t: FloatList):
        """This call method is used to evaluate the model.

        Parameters
        ----------
        x : :class:`~numpy.ndarray`
            The current x-axis meshgrid.

        y : :class:`~numpy.ndarray`
            The current y-axis meshgrid.

        
        z : :class:`~numpy.ndarray`
            The current z-axis meshgrid.
        
        t : :obj:`List`[:obj:`float`]
            The current time domain array.

        Returns
        -------
        :class:`~numpy.ndarray`
            The calculated concentration values.

        """
        self.get_grid_shape(x)
        self.assign_grids(x, y, z, t)

        self.rv = array(self.zero_arrays())

        for mode in self.modes:
            self.sources = getattr(self.settings.modes, mode).sources
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                getattr(self, f"{mode}")()

        return self.rv
    
    def assign_grids(self, x: ndarray, y: ndarray,  z: ndarray, t: FloatList) -> None:
        """Assign the meshgrids to the relevant attributes.

        Parameters
        ----------
        x : :class:`~numpy.ndarray`
            The current x-axis meshgrid.

        y : :class:`~numpy.ndarray`
            The current y-axis meshgrid.

        
        z : :class:`~numpy.ndarray`
            The current z-axis meshgrid.
        
        t : :obj:`List`[:obj:`float`]
            The current time domain array.
        
        Returns
        -------
        None

        """
        self.x = x
        self.y = y
        self.z = z
        self.t = t
    
    def get_grid_shape(self, grid: ndarray) -> Tuple[int]:
        """Returns the shape of the passed grid.

        Parameters
        ----------
        grid : :class:`~numpy.ndarray`
            The grid to be assessed.
        
        Returns
        -------
        :class:`Tuple`[:obj:`int`]
            The shape of the grid.

        """
        self.shape = grid.shape
    
    def get_cartesian_index_space(self):
        """Generate a cartesian product set of all grid indices.

        Returns
        -------
        :class:`Iterable`[:class:`Tuple`[:obj:`int`]]
            An iterable of tuples of indices.

        """

        return product(*[range(d) for d in self.shape])

    def get_raw_values(self, idx: int, idy: int, idz: int) -> Tuple[float]:
        """For a given set of grid indices, return raw values from the meshes.

        Parameters
        ----------
        idx : :obj:`int`
            The x index.

        idy : :obj:`int`
            The y index.

        idz : :obj:`int`
            The z index.

        Returns
        -------
        :class:`Tuple`[:obj:`float`]
            The x, y, and z values corresponding to the provided indices.
            
        """
        index = (idx, idy, idz)
        return self.x[index], self.y[index], self.z[index]

    @property
    def time(self):
        """Returns a :mod:`tqdm` iterable over the :attr:`time` iterable.

        Returns
        -------
        :obj:`Iterable`[:class:`Tuple`[:obj:`int`, :obj:`float`]
            The :mod:`tqdm` iterable over the :attr:`time` iterable
            
        """
        return tqdm(enumerate(self.t), total=len(self.t), **bar_args)

    def romberg(self, time: float, source: Source, idx: int, idy: int, idz: int) -> float:
        """ Performs romberg integration at the given grid location.

        This is performed between zero and the provided time.

        Parameters
        ----------
        time : :obj:`float`
            The time to integrate to.

        source : :class:`~.Source`    
            The source term in question.

        idx : :obj:`int`
            The x index.

        idy : :obj:`int`
            The y index.

        idz : :obj:`int`
            The z index.

        Returns
        -------
        :obj:`float`
            The computed concentration.

        """
        x, y, z = self.get_raw_values(idx, idy, idz)
        integrand = lambda tin: self.conc(source, x, y, z, tin)
        return romberg(integrand, 1e-100, time, tol=1e-100)
    
    def pointwise(self, source: Source, time: float) -> ndarray:
        """Evaluates the equation pointwise at every location in the meshgrids.

        Parameters
        ----------
        time : :obj:`float`
            The time to integrate to.

        source : :class:`~.Source`    
            The source term in question.

        Returns
        -------
        :class:`~numpy.ndarray`
            The array containing the computed concentrations.
        """
        return self.conc(source, self.x, self.y, self.z, time)

    def evaluate(self, rtime: float, idt: int, conc: List[ndarray], source: Source):
        """Calls the relevant method for evaluating the equations.
    
        Different methods are called depending on the integration method
        selected. 

        Parameters
        ----------
        rtime : :obj:`float`
            The time relative to the source release/start time.

        idt : :obj:`int`
            The absolute time index.

        conc : :class:`List`[:class:`~numpy.ndarray`]
            The list of grids to store the computed values in.

        source : :class:`~.Source`
            The source being evaluated.
            
        Returns
        -------
        None

        """

        if self.settings.integration_method == "romberg":
            for item in self.get_cartesian_index_space():
                conc[idt][item] += source.rate * self.romberg(rtime, source, *item)
        else:
            conc[idt] += source.rate * self.pointwise(source, rtime)
    
    def process(self, conc: List[ndarray]) -> ndarray:
        """Perform the integration method dependent post processing. 

        Parameters
        ----------
        conc : :class:`List`[:class:`~numpy.ndarray`]
            The list of grids to store the computed values in.

        Returns
        -------
        :class:`~numpy.ndarray`
            The final grids to be added to the total.

        """
        if self.settings.integration_method == "romberg":
            return array(conc)
        else:
            return cumtrapz(array(conc), **self.cumtrapz_kwargs)
    
    def log_start(self, name: str, id: str) -> None:
        """Print a log message the evaluation of a grid has started.

        Parameters
        ----------
        name : :obj:`str`
            The type of source.

        id : :obj:`str`
            The source id string.
        
        Returns
        -------
        None

        """
        print(f"Evaluating {name} source (id: {id}) for each time...") 

    def instantaneous(self): 
        """Evaluate all instanteneous sources.

        Returns
        -------
        None

        """
        for id, source in self.sources.items():
            conc = self.zero_arrays()
            self.log_start("instanteneous", id)
            for idt, time in self.time:
                stime = time - source.time
                if stime > 0:
                    conc[idt] += source.mass * self.pointwise(source, stime)
            self.rv += array(conc)
 
    def infinite_duration(self):
        """Evaluate all infinite duration sources.

        Returns
        -------
        None

        """
        for id, source in self.sources.items():
            conc = self.zero_arrays()
            self.log_start("infinite duration", id)
            for idt, time in self.time:
                stime = time - source.time
                if stime > 0:
                    self.evaluate(stime, idt, conc, source)
            self.rv += self.process(conc)
    
    def fixed_duration(self):
        """Evaluate all fixed duration sources.

        Returns
        -------
        None

        """
        for id, source in self.sources.items():
            conc = self.zero_arrays()
            conc_decay = self.zero_arrays()
            self.log_start("fixed duration", id)
            for idt, time in self.time:
                stime = time - source.start_time
                etime = time - source.start_time - source.end_time
                if stime > 0:
                    self.evaluate(stime, idt, conc, source)
                if etime > 0:
                    self.evaluate(etime, idt, conc_decay, source)
            self.rv += self.process(conc) - self.process(conc_decay)
    
    def conc(self, source: Source, x: Value, y: Value, z: Value, t: float) -> Value:
        """Evaluate various the model at a given location, time and source.

        Parameters
        ----------
        source : :class:`~.Source`
            The source being evaluated.

        x : :class:`~.Value`
            The x value.

        y : :class:`~.Value`
            The y value.

        z : :class:`~.Value`
            The z value.

        t : :obj:`float`
            the time value

        Returns
        -------
        :class:`~.Value`
            The calculated concentration.
        """
        r_x = self.exp(x, t, self.dim.x, source.x)
        r_y = self.exp(y, t, self.dim.y, source.y)
        r_z = self.exp(z, t, self.dim.z, source.z)
        return self.coefficient(t) * r_x * r_y * r_z

    def exp(self, pos: Value, time: float, bound: float, spos: float) -> Value:
        """The sum of exponentials in the Eddy diffusion model.

        Parameters
        ----------
        pos : :class:`~.Value`
            The position.

        t : :obj:`float`
            The time.

        bound : :obj:`float`
            The upper spatial bound.

        spos : :obj:`float`
            The source position.

        Returns
        -------
        :class:`~.Value`
            The calculated value.
        
        """

        image_setting = self.settings.models.eddy_diffusion.images

        term = lambda x: exp(-power(x, 2) / (4 * self.diff_coeff * time))
        exp_arg = lambda x: pos + 2 * x * bound
        image = lambda x: term(exp_arg(x) - spos) + term(exp_arg(x) + spos)

        image_index = 0

        rv = 0.0 if type(pos) is float else zeros(pos.shape)

        rv += image(image_index)

        if image_setting.mode == "manual":
            if image_setting.quantity:
                image_index = image_setting.quantity
                for idx in range(1, image_index + 1):
                    rv += image(idx) + image(-idx)
        else:
            while image_index < MAX_IMAGE:
                image_index += 1
                new_term = 0.0 if type(pos) is float else zeros(pos.shape)
                new_term += image(image_index) + image(-image_index)
                if self.geometric_variance(rv, rv + new_term) < 1 + 1e-10:
                    rv += new_term
                    break
                rv += new_term
        return rv
    
    def geometric_variance(self, old: Value, new: Value) -> float:
        """Calculate the geometric variance between two grids (or numbers)

        Parameters
        ----------
        old : :class:`~.Value`
            The old values.

        new : :class:`~.Value`
            The new values.

        Returns
        -------
        :obj:`float` 
            The geometric variance.

        """
        return exp(nanmean(square(log(old)-log(new))))

    def coefficient(self, t: float) -> float:
        """Computes the temporal decay coefficient.

        Parameters
        ----------
        t : :obj:`float` 
            The current time.

        Returns
        -------
        :obj:`float`
            The coefficient.

        """
        fa_rate = self.settings.fresh_air_flow_rate
        num = exp(-t * fa_rate / self.volume)
        den = 8 * power(pi * self.diff_coeff * t, 3.0 / 2.0)
        return num / den

    def diffusion_coefficient(self):
        """Computes the diffusion coefficient.

        Returns
        -------
        :obj:`float`
            The selected diffusion coefficient.

        """
        tkeb_settings = self.settings.models.eddy_diffusion.coefficient.tkeb
        vent_number = tkeb_settings.number_of_supply_vents
        air_flow_rate = tkeb_settings.total_air_flow_rate
        coeff = self.settings.models.eddy_diffusion.coefficient
        bound = tkeb_settings.bound

        if coeff.calculation == "EXPLICIT":
            return coeff.value
        else:
            tkeb_term = air_flow_rate /\
                power(self.volume * power(vent_number, 2), 1.0 / 3.0)
            if bound == "upper":
                return 0.827 * tkeb_term + 0.0565
            elif bound == "regression":
                return 0.824 * tkeb_term
            elif bound == "lower": 
                return max(0.822 * tkeb_term - 0.0565, 0.001)

    def zero_arrays(self):
        """Creates new time series lists of grids filed with zeros.

        Returns
        -------
        :class:`List`[:class:`~numpy.ndarray`]
            The list of grids to store the computed values in.

        """
        return [zeros(self.shape) for i in range(self.settings.time_samples)]



from tqdm import tqdm

from numpy import max
from numpy import linspace

from ridt.plot import PointPlot
from ridt.plot import LinePlot
from ridt.plot import ContourPlot

from ridt.data import DirectoryAgent

from .datastore import DataStore

from ridt.config import RIDTConfig
from ridt.config import Units

from ridt import bar_args


class DataStorePlotter:
    """The class which plots the requested data from a :class:`~.DataStore`.

    """

    geometries = {
        "points": PointPlot,
        "lines": LinePlot,
        "planes": ContourPlot
    }

    def __init__(self,
                 dir_agent: DirectoryAgent,
                 data_store: DataStore,
                 settings: RIDTConfig,
                 quantity: str) -> None:
        """The :class:`DataStorePlotter` constructor.

        Parameters
        ----------
        settings : :class:`~.RIDTConfig`
            The settings for the run in question.

        dir_agent : :class:`~.DirectoryAgent`
            The path to the output directory for the run.
        
        data_store : :class:`~.DataStore`
            The data store to be analysed.

        quantity: :obj:`str`
            The string id for the quantity stored in the data  store.

        """
        units = Units(settings)
        factor = getattr(units, f"{quantity}_factor")

        for geometry, plotter in DataStorePlotter.geometries.items():
            # Get the Plotter Settings object.
            config = getattr(settings.models.eddy_diffusion, f"{geometry}_plots")
            # Verify that plots are requested.
            if not config.output: continue
            print(f"Plotting {geometry} monitor locations...")
            # Generates the time indices to plot from the config object.
            try:
                indices = self.spread(settings.time_samples, config.number)
            except AttributeError:
                # This happens if it is a point.
                indices = None
            dir_agent.create_plot_dir(geometry, quantity)
            plotter = plotter(settings, dir_agent.pdir, quantity)
            # Loop over all items in this geometry class.
            for id, data in getattr(data_store, geometry).items():
                # Rescale data from SI units back into output units.
                rescaled_data = data / factor
                # Find the maximum value.
                max_val = max(rescaled_data)
                if geometry == "points":
                    plotter(id, rescaled_data)
                else:
                    # Loop over all requested time indices.
                    for idx in tqdm(indices, **bar_args):
                        plotter(id, rescaled_data[idx], max_val, idx)

    def spread(self, time_samples: int, number_of_plots: int):
        """Returns a spread of time indices over time domain.

        Parameters
        ----------
        time_samples : :obj:`int`
            The total number of time samples.

        number_of_plots : :obj:`int`
            The number of requested plots.

        Returns
        -------
        :obj:`List`[:obj:`float`]
            The time indices to be plotted.
        
        """
        if number_of_plots > time_samples:
            number_of_plots = time_samples
        return [int(i) for i in linspace(0, time_samples - 1, number_of_plots)]



import csv

from itertools import product

from os.path import join

from tqdm import tqdm

from numpy import ndarray

from ridt.base import RIDTOSError

from ridt.config import RIDTConfig
from ridt.config import Units

from .directoryagent import DirectoryAgent

from ridt.container import Domain

from .datastore import DataStore

from ridt import bar_args


class DataStoreCSVWriter:
    """Class that writes a :class:`~.DataStore` instance to CSV files.

    Attributes
    ----------
    settings : :class:`~.RIDTConfig`
        The settings for the run in question.

    dir_agent : :class:`~.DirectoryAgent`
        The path to the output directory for the run.
    
    domain : :class:`~.Domain`
        The instance of :class:`~.Domain` corresponding to :attr:`setting`.

    units : :class:`~.Units`
        The instance of :class:`~.Units` corresponding to :attr:`setting`.

    data_store : :class:`~.DataStore`
        The data store to be analysed.

    quantity: :obj:`str`
        The string id for the quantity stored in the data  store.

    """ 

    def __init__(self,
                 setting: RIDTConfig,
                 data_store: DataStore,
                 dir_agent: DirectoryAgent,
                 quantity: str):
        """The :class:`DataStoreCSVWriter` constructor.

        Parameters
        ----------
        settings : :class:`~.RIDTConfig`
            The settings for the run in question.

        dir_agent : :class:`~.DirectoryAgent`
            The path to the output directory for the run.
        
        data_store : :class:`~.DataStore`
            The data store to be analysed.

        quantity: :obj:`str`
            The string id for the quantity stored in the data  store.

        """
        self.dir_agent = dir_agent
        self.setting = setting
        self.units = Units(setting)
        self.domain = Domain(setting)
        self.quantity = quantity
        self.write(data_store)
    
    @property
    def geometries(self):
        """:obj:`list` [:obj:`str`] : the list of geometries selected for
        evaluation in :attr:`settings`.

        """
        locations = self.setting.models.eddy_diffusion.monitor_locations
        return [g for g, e in locations.evaluate.items() if e]

    def write(self, data_store: DataStore) -> None:
        """Loops over entries in the data store and writes the data to csv file.

        Parameters
        ----------
        data_store : :class:`~.DataStore`
            The data store to be analysed.
        
        Returns
        -------
        None

        """
        for geometry in self.geometries:
            self.dir_agent.create_data_dir(geometry, self.quantity)
            for id in getattr(data_store, geometry):
                self.write_csv(geometry, id, data_store.get(geometry, id))
    
    def write_csv(self, geometry: str, id: str, data: ndarray) -> None:
        """Takes string identifiers and the grid and writes them to a csv file.

        Parameters
        ----------
        geometry : :obj:`str`
            The type of grid to be written.

        id : :obj:`str`
            The id of the grid to be written.

        data : :class:`~numpy.ndarray`
            The grid to be written.

        Raises
        ------
        :class:`~.RIDTOSError`
            If unable to create the file on disk.
        
        Returns
        -------
        None

        """
        path = join(self.dir_agent.ddir, id + ".csv")
        factor = getattr(self.units, f"{self.quantity}_factor")

        try:
            f = open(path, 'w', newline="")
        except OSError as e:
            raise RIDTOSError(e)

        writer = csv.writer(f, delimiter=",")
        writer.writerow([
            f"time ({self.units.time})",
            f"x ({self.units.space})",
            f"y ({self.units.space})",
            f"z ({self.units.space})",
            f"value ({getattr(self.units, f'{self.quantity}')})"
        ])

        indices = list(product(*[range(i) for i in data.shape]))

        print(f"Writing {id} {self.quantity} data to a csv file...")
        for index in tqdm(indices, total=len(indices), **bar_args):
            values = self.domain.values(id, index)
            writer.writerow(list(values) + [data[index] / factor])
        f.close()


import unittest
import os
from os import listdir
from os.path import join
from os import remove
import shutil
from pathlib import Path

from ridt.config import ConfigFileParser

from ridt.container import Domain
from ridt.container.eddydiffusionrun import EddyDiffusionRun

from ridt.data import BatchDataStore
from ridt.data import DataStoreReader


class ST26(unittest.TestCase):

    """System Test 26. Test the system can
       output all computed quantities specified
       by the user requirements to the output directory."""

    def setUp(self) -> None:

        this_dir = os.path.dirname(os.path.abspath(__file__))
        with ConfigFileParser(join(this_dir, "st26/config.json")) as cfp:
            self.c = cfp

        self.bds = BatchDataStore()
        self.domain = Domain(self.c)
        self.output_dir = join(this_dir, "st26/data")
        Path(self.output_dir).mkdir(parents=True, exist_ok=True)
        self.edr = EddyDiffusionRun(self.c, self.output_dir)

    def tearDown(self) -> None:
        shutil.rmtree(self.output_dir)

    def test_verify(self):
        output_types = ["concentration", "exposure"]
        geometries = ["points", "lines", "planes"]

        plots = ["points_plots", "lines_plots", "planes_plots"]
        plot_output = []
        ed = self.c.models.eddy_diffusion
        for plot in plots:
            output = getattr(ed, plot).output
            plot_output.append(output)

        for output in output_types:
            for idx, geometry in enumerate(geometries):
                if plot_output[idx]:
                    _path = join(self.output_dir, geometry)
                    self.assertIn(output, listdir(_path))


if __name__ == "__main__":
    unittest.main()



import unittest
import json
import os

import numpy as np

from ridt.config import RIDTConfig

from ridt.container import Domain


class TestDomain(unittest.TestCase):

    def setUp(self) -> None:

        this_dir = os.path.dirname(os.path.abspath(__file__))
        path = os.path.join(this_dir, "../../default/config.json")
        with open(path) as f:
            loaded_json = json.load(f)

        self.config = RIDTConfig(loaded_json)
        self.domain = Domain(self.config)

    def test_point(self):

        """Tests to make sure that the types of the points
        are arrays."""

        points = self.config.models.eddy_diffusion.monitor_locations.points
        for point in points.values():
            for array in self.domain.points(point):
                self.assertEqual(type(array), np.ndarray)

    def test_line(self):

        """Tests to make sure that the types of the lines
        are arrays."""

        lines = self.config.models.eddy_diffusion.monitor_locations.lines
        for line in lines.values():
            for array in self.domain.lines(line):
                self.assertEqual(type(array), np.ndarray)

    def test_plane(self):

        """Tests to make sure that the types of the planes
        are arrays."""

        planes = self.config.models.eddy_diffusion.monitor_locations.planes
        for plane in planes.values():
            for array in self.domain.planes(plane):
                self.assertEqual(type(array), np.ndarray)



import csv

from io import TextIOWrapper

from os.path import join

from typing import List
from typing import Iterable
from typing import Type

from ridt.base import RIDTOSError

from ridt.config import RIDTConfig
from ridt.config import Units
from ridt.config import summary

from ridt.container import Domain

from ridt.equation import EddyDiffusion

from ridt.data import DirectoryAgent

from .datastoreanalyser import DataStoreAnalyser

from .resultcontainers import ResultContainer

StringList = List[str]


class ResultsWriter:
    """The Results Writer class.

    Iterates through all selected geometries and writes all quantities to disk.

    Attributes
    ----------
    setting : :class:`~.RIDTConfig`
        The settings for the run in question.
    
    domain : :class:`~.Domain`
        The instance of :class:`~.Domain` corresponding to :attr:`setting`.

    analysis : :class:`~.DataStoreAnalyser`
        :class:`~.DataStoreAnalyser` instance to be written.

    quantity: :obj:`str`
        The string id for the quantity stored in the data  store.

    thresholds : :obj:`list` [:obj:`float`]
        The threshold values corresponding to :attr:`quantity` defined in
        :attr:`setting`.
    
    units : :class:`~.Units`
        The instance of :class:`~.Units` corresponding to :attr:`setting`.

    dir_agent: :class:`~.DirectoryAgent`
        The directory agent for run.

    """ 
    def __init__(self,
                 setting: RIDTConfig,
                 analysis: DataStoreAnalyser,
                 dir_agent: DirectoryAgent,
                 quantity: str):
        """The :class`~.ResultsWriter` class initialiser.

        Parameters
        ----------
        setting : :class:`~.RIDTConfig`
            The settings for the run in question.

        analysis : :class:`~.DataStoreAnalyser`
            :class:`~.DataStoreAnalyser` instance to be written.

        dir_agent: :class:`~.DirectoryAgent`
            The directory agent for run.

        quantity : :obj:`str`
            The string id for the quantity stored in the data  store.

       """
        self.setting = setting
        self.units = Units(setting)
        self.quantity = quantity
        self.thresholds = self.threshold_converter()
        self.analysis = analysis
        self.dir_agent = dir_agent 
        self.domain = Domain(setting)
        if quantity == "concentration":
            self.summary()
        self.maximum()
        self.exceedance_analysis()
        self.extrema()

    @property
    def geometries(self):
        """:obj:`list` [:obj:`str`] : the list of geometries selected for
        evaluation in :attr:`setting`.

        """
        locations = self.setting.models.eddy_diffusion.monitor_locations
        return [g for g, e in locations.evaluate.items() if e]

    def threshold_converter(self):
        """Converts the threshold into SI units.

        Converts the thresholds in :attr:`setting` corresponding to
        :attr:`quantity` into SI units.

        Returns
        -------
        :obj:`list` [:obj:`float`]
            The list of threshold values in SI units. 
        """
        tld = [t.value for t in getattr(self.setting.thresholds, self.quantity)]
        return getattr(self.units, f"{self.quantity}_converter")(tld)

    def write(self, file_path: str, header: StringList, lines: Iterable) -> None:
        """Write array of values to csv file.

        Parameters
        ----------
        file_path : :obj:`str`
            The path to save the data to.
        header : :obj:`list` [:obj:`str`]
            The header for the csv file.
        lines : Iterable
            The array of values to be written to the csv file.

        Returns
        -------
        None

        Raises
        ------
        :class:`~.RIDTOSError`
            If cannot create file on disk.

        """
        try:
            f = open(file_path, 'w', newline="")
        except OSError as e:
            raise RIDTOSError(e)

        writer = csv.writer(f, delimiter=",")
        writer.writerow(header)
        for line in lines:
            writer.writerow(line)
        f.close()

    def get_valid(self, geometry: str, results: List[Type[ResultContainer]]):
        """Filter valid results for a given geometry from list of results.

        Parameters
        ----------
        geometry : :obj:`str`
            The geometry in to filter for.

        results : :obj:`list` [:class:`~.ResultContainer`]
            The list of :class:`~.ResultConainer`.

        Returns
        -------
        :obj:`list` [:class:`~.ResultContainer`]
            The filtered list of :class:`~.ResultConainer`.
        
        """
        return [i for i in results if i.geometry == geometry and i.valid]
    
    def maximum(self):
        """Write the relevant :class:`~.Maximum` results to file.
    
        Loops through all geometries and writes the Maximum results to
        a csv file.

        Returns
        -------
        None
        
        """ 
        for geometry in self.geometries:
            self.dir_agent.create_analysis_dir(geometry, self.quantity)
            items = self.get_valid(geometry, self.analysis.maximum)
            items.sort(reverse=True)
            rows = [item.row for item in items]
            if rows:
                path = join(self.dir_agent.adir, items[0].fname)
                self.write(path, items[0].header, rows)

    def exceedance_analysis(self):
        """Write the relevant exceedance results to file.
    
        Loops through all geometries and writes the :class:`~.Exceedance`,
        :class:`~.PercentExceedance`, and :class:`~.MaxPercentExceedance`
        results to a csv file.

        Returns
        -------
        None
        
        """
        results = [
            (self.analysis.exceedance, False),
            (self.analysis.percent_exceedance, False),
            (self.analysis.max_percent_exceedance, True)
        ]
        for r, reverse in results:
            for geometry in self.geometries:
                self.dir_agent.create_analysis_dir(geometry, self.quantity)
                for t in self.thresholds:
                    valid = self.get_valid(geometry, r)
                    items = [i for i in valid if i.threshold == t]
                    items.sort(reverse=reverse)
                    rows = [item.row for item in items]
                    if rows:
                        path = join(self.dir_agent.adir, items[0].fname)
                        self.write(path, items[0].header, rows)
    
    def extrema(self):
        """Computes and writes extrema results to disk.

        loops through all items in :attr:`analysis` and for each geometry type
        writes the following information to the output file:

        The maximum value.

        The fastest time to all threshold values.

        The fastest time to the defined percentage of domain to all threshold
        values.

        The largest percentage that exceeds all threshold values.
        The 

        Raises
        ------
        :class:`~.RIDTOSError`
            If unable to create the output file on disk.

        """
        try:
            fname = f"{self.quantity}_extrema.txt"
            f = open(join(self.dir_agent.outdir, fname), 'w')
        except OSError as e:
            raise RIDTOSError(e)

        self.title(f, self.analysis.maximum[0].title)
        for geometry in self.geometries:
            items = self.get_valid(geometry, self.analysis.maximum)
            if items:
                self.subtitle(f, items[0].extreme_title)
                item = max(items)
                f.write(item.string)

        results = [
            (self.analysis.exceedance, False),
            (self.analysis.percent_exceedance, False),
            (self.analysis.max_percent_exceedance, True)
        ]

        for r, reverse in results:
            self.title(f, r[0].title)
            for t in self.thresholds:
                for geometry in self.geometries:
                    valid = self.get_valid(geometry, r)
                    items = [i for i in valid if i.threshold == t]
                    if items:
                        self.subtitle(f, items[0].extreme_title)
                        item = min(items) if not reverse else max(items)
                        f.write(item.string)
        f.close()

    def title(self, file: TextIOWrapper, title: str):
        """Writes a formatted title string to a file object.

        Parameters
        ----------
        file : :obj:`TextIOWrapper` 
            The file object to write to.

        title : :obj:`str`
            The title string.

        Returns
        -------
        None

        """
        file.write("\n")
        file.write("".join("=" for i in range(len(title))) + "\n")
        file.write(title + "\n")
        file.write("".join("=" for i in range(len(title))) + "\n")

    def subtitle(self, file: TextIOWrapper, subtitle: str):
        """Writes a formatted subtitle string to a file object.

        Parameters
        ----------
        file : :obj:`TextIOWrapper` 
            The file object to write to.

        subtitle : :obj:`str`
            The subtitle string.
        
        Returns
        -------
        None
        """
        file.write("\n")
        file.write("".join("-" for i in range(len(subtitle))) + "\n")
        file.write(subtitle + "\n")
        file.write("".join("-" for i in range(len(subtitle))) + "\n")

    def summary(self):
        """Saves a run summary to a text file in the run output directory.

        Raises
        ------
        :class:`~.RIDTOSError`
            If unable to create the output file on disk.

        """
        try:
             f = open(join(self.dir_agent.outdir, "run_summary.txt"), 'w')
        except OSError as e:
            raise RIDTOSError(e)
    
        char_diff = self.characteristic_diffusion_time
        f.write("Characteristic diffusion times:\n")
        f.write(f"\tx: {char_diff['x']:.2f}{self.setting.time_units}\n")
        f.write(f"\ty: {char_diff['y']:.2f}{self.setting.time_units}\n")
        f.write(f"\tz: {char_diff['z']:.2f}{self.setting.time_units}\n")
        f.write(f"\tV^(1/3): {char_diff['v']:.2f}{self.setting.time_units}\n")
        f.write("\n")
        if self.setting.models.eddy_diffusion.monitor_locations.evaluate["domain"]:
            ttwm = self.analysis.time_to_well_mixed
            if ttwm:
                value = f"{ttwm:.2f}{self.setting.time_units}\n"
            else:
                value = "not within lifetime of simulation\n"
        else:
            value = "domain data not available\n"
        f.write(f"Time to well mixed: {value}")
        f.write("\n")
        f.write(summary(self.setting))
        f.close()

    @property
    def characteristic_diffusion_time(self):
        """:obj:`dict` [:obj:`str`, :obj:`float`] : the characteristic 
        diffusion times for all dimensions.

        """
        solver = EddyDiffusion(self.setting)
        dim = self.setting.dimensions
        l = pow(dim.x * dim.y * dim.z, 1.0 / 3.0)
        return {
            "x": dim.x * dim.x / solver.diff_coeff,
            "y": dim.y * dim.y / solver.diff_coeff,
            "z": dim.z * dim.z / solver.diff_coeff,
            "v": l * l / solver.diff_coeff
        }

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass
   


import unittest
import os
from os import listdir
from os.path import join
from os import remove
import shutil
from pathlib import Path

from ridt.config import ConfigFileParser

from ridt.container import Domain
from ridt.container.eddydiffusionrun import EddyDiffusionRun

from ridt.data import BatchDataStore


class ST20(unittest.TestCase):

    """System Test 20. Test the system can
       produce contour plots of data where requested. """

    def setUp(self) -> None:

        this_dir = os.path.dirname(os.path.abspath(__file__))
        with ConfigFileParser(join(this_dir, "st20/config.json")) as cfp:
            self.c = cfp

        self.bds = BatchDataStore()
        self.domain = Domain(self.c)
        self.output_dir = join(this_dir, "st20/plots")
        Path(self.output_dir).mkdir(parents=True, exist_ok=True)
        self.edr = EddyDiffusionRun(self.c, self.output_dir)

    def tearDown(self) -> None:
        shutil.rmtree(self.output_dir)

    def test_verify(self):
        output_types = ["concentration", "exposure"]
        num = self.c.models.eddy_diffusion.planes_plots.number
        for output in output_types:
            _path = join(self.output_dir, "planes", output, "plots")
            for name, val in self.c.models.eddy_diffusion.monitor_locations.planes.items():
                for i in range(num):
                    val = i * (self.c.total_time / (num - 1))
                    self.assertIn(f"{name}-{val}0s.png", listdir(_path))


if __name__ == "__main__":
    unittest.main()



import unittest
import os

from ridt.config import RIDTConfig
from ridt.config.configfileparser import ConfigFileParser
from ridt.config.configfileparser import ConfigFileParserOSError
from ridt.config.configfileparser import ConfigFileParserJSONError
from ridt.config.configfileparser import ConfigFileParserValidationError


class TestConfigFileParser(unittest.TestCase):

    """The unit tests for the :class:`~.ConfigFileParser` class."""

    def setUp(self) -> None:

        """setUp method which instantiates the :class:`~.RIDTConfig` class,
        and the :class:`~.ConfigFileParser` class."""

        self.this_dir = os.path.dirname(os.path.abspath(__file__))
        self.ridt = RIDTConfig

    def test_parse_file(self):

        """Checks if the :class:`~.ConfigFileParserOSError` error
        is passed when the path is unknown.
        Checks to see if the :class:`~.ConfigFileParserJSONError` error
        is passed when there isa decode error in the json file."""

        with self.assertRaises(ConfigFileParserOSError):
            ConfigFileParser("random_path")

        with self.assertRaises(ConfigFileParserJSONError):
            ConfigFileParser(os.path.join(self.this_dir, "test_resources/decode_error_config.json"))

    def test_instantiate_settings(self):

        """"Checks to make sure that the :class:`~.ConfigFileParserValidationError`
        error passes when there is a missing input within the config file."""

        with self.assertRaises(ConfigFileParserValidationError):
            ConfigFileParser(os.path.join(self.this_dir, "test_resources/missing_input_config.json"))


if __name__ == "__main__":
    unittest.main()
